@@ -87,7 +87,7 @@ class ChatModel(QAbstractTableModel):
             elif role == Qt.DisplayRole and index.column() == 2:
                 result = toQVariant(m18n(chatLine.message))
             elif role == Qt.ForegroundRole and index.column() == 2:
-                palette = KApplication.palette()
+                palette = KApplication.palette() # pylint: disable=no-member
                 color = 'blue' if chatLine.isStatusMessage else palette.windowText(
                 )
                 result = toQVariant(QColor(color))
@@ -196,7 +196,6 @@ class ChatWindow(QWidget):
 
     def leave(self):
         """leaving the chat"""
-        # TODO: send "left" message
         self.hide()
 
     def receiveLine(self, chatLine):
