@@ -26,26 +26,26 @@ signal.signal(signal.SIGINT, signal.SIG_DFL)
 
 import os
 import sys
-import csv
 import subprocess
 import random
 import shutil
 import time
 import gc
-from tempfile import mkdtemp
 
 from optparse import OptionParser
 
-from common import Debug, StrMixin
+from common import Debug, StrMixin, isPython3, nativeString
 from util import removeIfExists, gitHead, checkMemory
+from compat import Csv, CsvWriter
 
 # fields in row:
 RULESETFIELD = 0
 AIFIELD = 1
 COMMITFIELD = 2
-GAMEFIELD = 3
-TAGSFIELD = 4
-PLAYERSFIELD = 5
+PYTHON23FIELD = 3
+GAMEFIELD = 4
+TAGSFIELD = 5
+PLAYERSFIELD = 6
 
 OPTIONS = None
 
@@ -53,45 +53,34 @@ OPTIONS = None
 class Clone(object):
 
     """make a temp directory for commitId"""
-    clones = {}
-
-    def __new__(cls, commitId):
-        if commitId in cls.clones:
-            return cls.clones[commitId]
-        return object.__new__(cls)
 
     def __init__(self, commitId):
         self.commitId = commitId
-        self.clones[commitId] = self
         if commitId is 'current':
             self.tmpdir = os.path.abspath('..')
             srcDir = os.path.join(self.tmpdir, 'src')
             assert os.path.exists(srcDir), '{} does not exist'.format(srcDir)
         else:
-            self.tmpdir = mkdtemp(prefix='kajonggtest.{}.'.format(commitId))
-            subprocess.Popen('git clone --local --no-checkout -q .. {temp}'.format(
-                temp=self.tmpdir).split()).wait()
-            subprocess.Popen('git checkout -q {commitId}'.format(
-                commitId=commitId).split(), cwd=self.tmpdir).wait()
-
-    def remove(self):
-        """remove my tmpdir"""
-        del self.clones[self.commitId]
-        if self.commitId != 'current' and '/tmp/' in self.tmpdir:
-            shutil.rmtree(self.tmpdir)
+            self.tmpdir = os.path.expanduser(os.path.join('~', '.kajongg', '.cache', commitId))
+            if not os.path.exists(self.tmpdir):
+                subprocess.Popen('git clone --shared --no-checkout -q .. {temp}'.format(
+                    temp=self.tmpdir).split()).wait()
+                subprocess.Popen('git checkout -q {commitId}'.format(
+                    commitId=commitId).split(), cwd=self.tmpdir).wait()
+                if os.path.islink('twisted'):
+                    cpCmd = ['ln', '-s', '{}/twisted'.format(os.getcwd()),
+                             '{}/src/twisted'.format(self.tmpdir)]
+                    subprocess.Popen(cpCmd).wait()
 
     @classmethod
-    def removeUnused(cls):
-        """remove clones we do not use anymore"""
-        for commitId in cls.clones.keys():
-            if not any(x.commitId == commitId for x in Server.servers):
-                cls.clones[commitId].remove()
+    def cloneDirectory(cls):
+        """the directory for this git commit"""
+        return os.path.expanduser(os.path.join('~', '.kajongg', '.cache'))
 
     @classmethod
-    def removeAll(cls):
-        """remove all clones even if they are in use"""
-        for cloneKey in cls.clones.keys()[:]:
-            cls.clones[cloneKey].remove()
+    def commitDirectory(cls, commitId):
+        """the source directory for this git commit"""
+        return os.path.join(cls.cloneDirectory(), commitId)
 
 
 class Client(object):
@@ -117,14 +106,19 @@ class Server(StrMixin):
 
     """represents a kajongg server instance. Called when we want to start a job."""
     servers = []
+    count = 0
 
     def __new__(cls, job):
         """can we reuse an existing server?"""
         running = Server.allRunningJobs()
         if len(running) >= OPTIONS.clients:
             raise TooManyClients
+        maxClientsPerServer = OPTIONS.clients / OPTIONS.servers
         matchingServers = list(
-            x for x in cls.servers if x.commitId == job.commitId)
+            x for x in cls.servers
+            if x.commitId == job.commitId
+            and x.pythonVersion == job.pythonVersion
+            and len(x.jobs) < maxClientsPerServer)
         if matchingServers:
             result = sorted(matchingServers, key=lambda x: len(x.jobs))[0]
         else:
@@ -146,7 +140,9 @@ class Server(StrMixin):
             self.jobs = []
             self.process = None
             self.socketName = None
+            self.portNumber = None
             self.commitId = job.commitId
+            self.pythonVersion = job.pythonVersion
             self.clone = Clone(job.commitId)
             self.start(job)
         else:
@@ -162,35 +158,34 @@ class Server(StrMixin):
         """start this server"""
         assert self.process is None, 'Server.start already has a process'
         self.jobs.append(job)
-        if OPTIONS.usePort:
-            self.socketName = random.randrange(1025, 65000)
-        else:
-            self.socketName = os.path.expanduser(os.path.join('~', '.kajongg',
-                'sock{id}.{rnd}'.format(id=id(self), rnd=random.randrange(10000000))))
         assert self.commitId == job.commitId
-        print('starting server for %s' % job)
-        cmd = [os.path.join(job.srcDir(), 'kajonggserver.py')]
-        if OPTIONS.server3:
-            cmd.insert(0, 'python3')
-        elif os.name == 'nt':
-            cmd.insert(0, 'python')
-        else:
-            cmd.insert(0, 'python2')
+        cmd = [os.path.join(
+            job.srcDir(),
+            'kajonggserver.py')]
+        cmd.insert(0, 'python{}'.format(self.pythonVersion))
         if OPTIONS.usePort:
-            cmd.append('--port={sock}'.format(sock=self.socketName))
+            self.portNumber = random.randrange(1025, 65000)
+            cmd.append('--port={port}'.format(port=self.portNumber))
         else:
+            Server.count += 1
+            self.socketName = os.path.expanduser(
+                os.path.join('~', '.kajongg',
+                             'sock{commit}.py{py}.{ctr}'.format(
+                                 commit=self.commitId, py=self.pythonVersion, ctr=Server.count)))
             cmd.append('--socket={sock}'.format(sock=self.socketName))
         if OPTIONS.debug:
             cmd.append('--debug={dbg}'.format(dbg=','.join(OPTIONS.debug)))
         if OPTIONS.qt5:
             cmd.append('--qt5')
         if OPTIONS.log:
-            self.process = subprocess.Popen(cmd, cwd=job.srcDir(),
+            self.process = subprocess.Popen(
+                cmd, cwd=job.srcDir(),
                 stdout=job.logFile, stderr=job.logFile)
         else:
             # reuse this server (otherwise it stops by itself)
             cmd.append('--continue')
             self.process = subprocess.Popen(cmd, cwd=job.srcDir())
+        print('{} started'.format(self))
 
     def stop(self, job=None):
         """maybe stop the server"""
@@ -202,17 +197,14 @@ class Server(StrMixin):
         if len(self.jobs) == 0:
             self.servers.remove(self)
             if self.process:
-                print(
-                    'killing server %s%s' %
-                    (self, ' for {}'.format(job) if job else ''))
                 try:
                     self.process.terminate()
                     _ = self.process.wait()
                 except OSError:
                     pass
-            if self.socketName and not OPTIONS.usePort:
+                print('{} killed'.format(self))
+            if self.socketName:
                 removeIfExists(self.socketName)
-        Clone.removeUnused()
 
     @classmethod
     def stopAll(cls):
@@ -225,14 +217,19 @@ class Server(StrMixin):
             server.stop()
 
     def __unicode__(self):
-        return u'{} pid={} sock={}'.format(self.commitId, self.process.pid, self.socketName)
+        return u'Server {} Python{}{} {}'.format(
+            self.commitId,
+            self.pythonVersion,
+            ' pid={}'.format(self.process.pid) if Debug.process else u'',
+            'port={}'.format(self.portNumber) if self.portNumber else 'socket={}'.format(self.socketName))
 
 
 class Job(StrMixin):
 
     """a simple container"""
 
-    def __init__(self, ruleset, aiVariant, commitId, game):
+    def __init__(self, pythonVersion, ruleset, aiVariant, commitId, game):
+        self.pythonVersion = pythonVersion
         self.ruleset = ruleset
         self.aiVariant = aiVariant
         self.commitId = commitId
@@ -245,31 +242,36 @@ class Job(StrMixin):
 
     def srcDir(self):
         """the path of the directory where the particular test is running"""
-        return os.path.join(Clone.clones[self.commitId].tmpdir, 'src')
+        return os.path.join(Clone.commitDirectory(self.commitId), 'src')
+
+    def __startProcess(self, cmd):
+        """call Popen"""
+        if OPTIONS.log:
+            self.process = subprocess.Popen(
+                cmd, cwd=self.srcDir(),
+                stdout=self.logFile, stderr=self.logFile)
+        else:
+            self.process = subprocess.Popen(cmd, cwd=self.srcDir())
+        print('       %s started' % (self))
 
     def start(self):
         """start this job"""
+        # pylint: disable=too-many-branches
         self.server = Server(self)
         # never login to the same server twice at the
         # same time with the same player name
         player = self.server.jobs.index(self) + 1
-        if OPTIONS.usePort:
-            socketArg = '--port={sock}'.format(sock=self.server.socketName)
-        else:
-            socketArg = '--socket={sock}'.format(sock=self.server.socketName)
         cmd = [os.path.join(self.srcDir(), 'kajongg.py'),
-              '--game={game}'.format(game=self.game),
-              '--socket={sock}'.format(sock=self.server.socketName),
-              '--player={tester} {player}'.format(
-                player=player,
-                tester=u'Tüster'.encode('utf-8')),
-              '--ruleset={ap}'.format(ap=self.ruleset)]
-        if OPTIONS.client3:
-            cmd.insert(0, 'python3')
-        elif os.name == 'nt':
-            cmd.insert(0, 'python')
-        else:
-            cmd.insert(0, 'python2')
+               '--game={game}'.format(game=self.game),
+               '--player={tester} {player}'.format(
+                   player=player,
+                   tester=nativeString(u'Tüster')),
+               '--ruleset={ap}'.format(ap=self.ruleset)]
+        if self.server.socketName:
+            cmd.append('--socket={sock}'.format(sock=self.server.socketName))
+        if self.server.portNumber:
+            cmd.append('--port={port}'.format(port=self.server.portNumber))
+        cmd.insert(0, 'python{}'.format(self.pythonVersion))
         if OPTIONS.rounds:
             cmd.append('--rounds={rounds}'.format(rounds=OPTIONS.rounds))
         if self.aiVariant != 'Default':
@@ -286,12 +288,7 @@ class Job(StrMixin):
             cmd.append('--playopen')
         if OPTIONS.debug:
             cmd.append('--debug={dbg}'.format(dbg=','.join(OPTIONS.debug)))
-        print('starting            %s' % self)
-        if OPTIONS.log:
-            self.process = subprocess.Popen(cmd, cwd=self.srcDir(),
-                stdout=self.logFile, stderr=self.logFile)
-        else:
-            self.process = subprocess.Popen(cmd, cwd=self.srcDir())
+        self.__startProcess(cmd)
         self.started = True
 
     def check(self, silent=False):
@@ -302,10 +299,7 @@ class Job(StrMixin):
         if result is not None:
             self.process = None
             if not silent:
-                print(
-                    '   Game over: %s%s' %
-                    ('Return code: %s ' %
-     result if result else '', self))
+                print('       {} done{}'.format(self, 'Return code: {}'.format(result) if result else ''))
             self.server.jobs.remove(self)
 
     @property
@@ -314,7 +308,7 @@ class Job(StrMixin):
         if self.__logFile is None:
             logDir = os.path.expanduser(
                 os.path.join('~', '.kajongg', 'log', str(self.game),
-                self.ruleset, self.aiVariant))
+                             self.ruleset, self.aiVariant, str(self.pythonVersion)))
             if not os.path.exists(logDir):
                 os.makedirs(logDir)
             logFileName = self.commitId
@@ -336,22 +330,24 @@ class Job(StrMixin):
         ruleset = self.shortRulesetName()
         aiName = u'AI={}'.format(
             self.aiVariant) if self.aiVariant != u'Default' else u''
-        commit = u'commit={}'.format(self.commitId)
-        return u' '.join([pid, game, ruleset, aiName, commit]).replace('  ', ' ')
+        return u' '.join([
+            self.commitId, 'Python{}'.format(self.pythonVersion), pid, game, ruleset, aiName]).replace('  ', ' ')
 
 
 def neutralize(rows):
     """remove things we do not want to compare"""
     for row in rows:
         for idx, field in enumerate(row):
-            if field.startswith('Tester '):
-                row[idx] = 'Tester'
+            field = field.replace(' ', '')
+            if field.startswith(u'Tester ') or field.startswith(u'Tüster'):
+                field = 'Tester'
             if 'MEM' in field:
                 parts = field.split(',')
                 for part in parts[:]:
                     if part.startswith('MEM'):
                         parts.remove(part)
-                row[idx] = ','.join(parts)
+                field = ','.join(parts)
+            row[idx] = field
         yield row
 
 KNOWNCOMMITS = set()
@@ -361,55 +357,64 @@ def onlyExistingCommits(commits):
     """filter out non-existing commits"""
     global KNOWNCOMMITS  # pylint: disable=global-statement
     if len(KNOWNCOMMITS) == 0:
-        for branch in subprocess.check_output('git branch'.split(), env={'LANG': 'C'}).split('\n'):
-            if not 'detached' in branch and not 'no branch' in branch:
+        for branch in subprocess.check_output(b'git branch'.split()).decode().split('\n'):
+            if 'detached' not in branch and 'no branch' not in branch:
                 KNOWNCOMMITS |= set(subprocess.check_output(
-                    'git log --max-count=200 --pretty=%h {branch}'.format(
-                        branch=branch[2:]).split()).split('\n'))
-    return list(x for x in commits if x in KNOWNCOMMITS)
+                    'git log --max-count=200 --pretty=%H {branch}'.format(
+                        branch=branch[2:]).split()).decode().split('\n'))
+    result = list()
+    for commit in commits:
+        if any(x.startswith(commit) for x in KNOWNCOMMITS):
+            result.append(commit)
+    return result
 
 
 def removeInvalidCommits(csvFile):
     """remove rows with invalid git commit ids"""
     if not os.path.exists(csvFile):
         return
-    rows = list(csv.reader(open(csvFile, 'rb'), delimiter=';'))
+    rows = list(Csv.reader(csvFile))
     _ = set(x[COMMITFIELD] for x in rows)
     csvCommits = set(
         x for x in _ if set(
             x) <= set(
-            '0123456789abcdef') and len(
-                x) >= 7)
+                '0123456789abcdef') and len(
+                    x) >= 7)
     nonExisting = set(csvCommits) - set(onlyExistingCommits(csvCommits))
     if nonExisting:
         print(
             'removing rows from kajongg.csv for commits %s' %
             ','.join(nonExisting))
-        writer = csv.writer(open(csvFile, 'wb'), delimiter=';')
+        writer = CsvWriter(csvFile)
         for row in rows:
             if row[COMMITFIELD] not in nonExisting:
                 writer.writerow(row)
-    # now remove all logs referencing obsolete commits
-    for dirName, _, fileNames in os.walk('log'):
+    # remove all logs referencing obsolete commits
+    logDir = os.path.expanduser(os.path.join('~', '.kajongg', 'log'))
+    for dirName, _, fileNames in os.walk(logDir):
         for fileName in fileNames:
-            fullName = os.path.join(dirName, fileName)
             if fileName not in KNOWNCOMMITS and fileName != 'current':
-                os.remove(fullName)
+                os.remove(os.path.join(dirName, fileName))
         try:
             os.removedirs(dirName)
         except OSError:
             pass  # not yet empty
 
+    # remove all clones for obsolete commits
+    for commitDir in os.listdir(Clone.cloneDirectory()):
+        if not any(x.startswith(commitDir) for x in KNOWNCOMMITS):
+            removeDir = os.path.join(Clone.cloneDirectory(), commitDir)
+            shutil.rmtree(removeDir)
 
 def readGames(csvFile):
     """returns a dict holding a frozenset of games for each variant"""
     if not os.path.exists(csvFile):
         return
-    allRows = neutralize(csv.reader(open(csvFile, 'rb'), delimiter=';'))
-    if not allRows:
+    allRowsGenerator = neutralize(Csv.reader(csvFile))
+    if not allRowsGenerator:
         return
     # we want unique tuples so we can work with sets
-    allRows = set(tuple(x) for x in allRows)
+    allRows = set(tuple(x) for x in allRowsGenerator)
     games = dict()
     # build set of rows for every ai
     for variant in set(tuple(x[:COMMITFIELD]) for x in allRows):
@@ -417,6 +422,37 @@ def readGames(csvFile):
             x for x in allRows if tuple(x[:COMMITFIELD]) == variant)
     return games
 
+def hasDifferences(rows):
+    """True if rows have unwanted differences"""
+    return (len(set(tuple(list(x)[GAMEFIELD:]) for x in rows))
+            > len(set(tuple(list(x)[:COMMITFIELD]) for x in rows)))
+
+def firstDifference(rows):
+    """reduce to two rows showing a difference"""
+    result = rows
+    last = rows[-1]
+    while hasDifferences(result):
+        last = result[-1]
+        result = result[:-1]
+    return list([result[-1], last])
+
+def closerLook(gameId, gameIdRows):
+    """print detailled info about one difference"""
+    for ruleset in OPTIONS.rulesets:
+        for intelligence in OPTIONS.allAis:
+            shouldBeIdentical = list(x for x in gameIdRows if x[RULESETFIELD] == ruleset and x[AIFIELD] == intelligence)
+            for commit in list(x[COMMITFIELD] for x in shouldBeIdentical):
+                rows2 = list(x for x in shouldBeIdentical if x[COMMITFIELD] == commit)
+                if hasDifferences(rows2):
+                    first = firstDifference(rows2)
+                    print('Game {} {} {} {} has differences between Python2 and Python3'.format(
+                        gameId, ruleset, intelligence, commit))
+            for py23 in '23':
+                rows2 = list(x for x in shouldBeIdentical if x[PYTHON23FIELD] == py23)
+                if hasDifferences(rows2):
+                    first = firstDifference(rows2)
+                    print('Game {} {} {} Python{} has differences between commits {} and {}'.format(
+                        gameId, ruleset, intelligence, py23, first[0][COMMITFIELD], first[1][COMMITFIELD]))
 
 def printDifferingResults(rowLists):
     """if most games get the same result with all tried variants,
@@ -430,21 +466,23 @@ def printDifferingResults(rowLists):
             allGameIds[rowId].append(row)
     differing = []
     for key, value in allGameIds.items():
-        if len(set(tuple(list(x)[GAMEFIELD:]) for x in value)) > len(set(tuple(list(x)[:COMMITFIELD]) for x in value)):
+        if hasDifferences(value):
             differing.append(key)
     if not differing:
         print('no games differ')
-    elif float(len(differing)) / len(allGameIds) < 0.20:
+    else:
         print(
-            'differing games (%d out of %d): %s' % (len(differing), len(allGameIds),
-             ' '.join(sorted(differing, key=int))))
-
+            'differing games (%d out of %d): %s' % (
+                len(differing), len(allGameIds),
+                ' '.join(sorted(differing, key=int))))
+        # now look closer at one example. Differences may be caused by git commits or by py2/p3
+        for gameId in sorted(differing):
+            closerLook(gameId, allGameIds[gameId])
 
 def evaluate(games):
     """evaluate games"""
     if not games:
         return
-    commonGames = set()
     for variant, rows in games.items():
         gameIds = set(x[GAMEFIELD] for x in rows)
         if len(gameIds) != len(set(tuple(list(x)[GAMEFIELD:]) for x in rows)):
@@ -456,39 +494,25 @@ def evaluate(games):
                     print(game, end=' ')
             print()
             break
-        commonGames &= gameIds
     printDifferingResults(games.values())
     print()
     print('the 3 robot players always use the Default AI')
     print()
-    print('common games:')
     print('{ruleset:<25} {ai:<20} {games:>5}     {points:>4}                      human'.format(
         ruleset='Ruleset', ai='AI variant', games='games', points='points'))
     for variant, rows in games.items():
         ruleset, aiVariant = variant
-        print(
-            '{ruleset:<25} {ai:<20} {games:>5}  '.format(ruleset=ruleset[:25], ai=aiVariant[:20],
-            games=len(commonGames)), end=' ')
+        print('{ruleset:<25} {ai:<20} {rows:>5}  '.format(
+            ruleset=ruleset[:25], ai=aiVariant[:20],
+            rows=len(rows)), end=' ')
         for playerIdx in range(4):
-            print('{p:>8}'.format(p=sum(int(x[PLAYERSFIELD + 1 + playerIdx * 4])
-                    for x in rows if x[GAMEFIELD] in commonGames)), end=' ')
-        print()
-    print()
-    print('all games:')
-    for variant, rows in games.items():
-        ruleset, aiVariant = variant
-        if len(rows) > len(commonGames):
             print(
-                '{ruleset:<25} {ai:<20} {rows:>5}  '.format(ruleset=ruleset[:25], ai=aiVariant[:20],
-                rows=len(rows)), end=' ')
-            for playerIdx in range(4):
-                print(
-                    '{p:>8}'.format(
-                        p=sum(
-                            int(x[
-                                PLAYERSFIELD + 1 + playerIdx * 4]) for x in rows)),
-                    end=' ')
-            print()
+                '{p:>8}'.format(
+                    p=sum(
+                        int(x[
+                            PLAYERSFIELD + 1 + playerIdx * 4]) for x in rows)),
+                end=' ')
+        print()
 
 
 def startingDir():
@@ -499,7 +523,7 @@ def startingDir():
 def getJobs(jobs):
     """fill the queue"""
     try:
-        while len(jobs) < OPTIONS.servers * OPTIONS.clients:
+        while len(jobs) < OPTIONS.clients:
             jobs.append(next(OPTIONS.jobs))
     except StopIteration:
         pass
@@ -554,49 +578,58 @@ def doJobs():
 def parse_options():
     """parse options"""
     parser = OptionParser()
-    parser.add_option('', '--gui', dest='gui', action='store_true',
+    parser.add_option(
+        '', '--gui', dest='gui', action='store_true',
         default=False, help='show graphical user interface')
-    parser.add_option('', '--qt5', dest='qt5', action='store_true',
+    parser.add_option(
+        '', '--23', dest='py23', action='store_true',
+        default=False, help='Use both Python 2 and 3. If you explicitly want to use only one of them'
+                            'use python2 kajonggtest or python3 kajonggtest')
+    parser.add_option(
+        '', '--qt5', dest='qt5', action='store_true',
         default=False, help='Force using Qt5')
-    parser.add_option('', '--ruleset', dest='rulesets', default='ALL',
+    parser.add_option(
+        '', '--ruleset', dest='rulesets', default='ALL',
         help='play like a robot using RULESET: comma separated list. If missing, test all rulesets',
         metavar='RULESET')
-    parser.add_option('', '--rounds', dest='rounds',
+    parser.add_option(
+        '', '--rounds', dest='rounds',
         help='play only # ROUNDS per game',
         metavar='ROUNDS')
-    parser.add_option('', '--ai', dest='aiVariants',
+    parser.add_option(
+        '', '--ai', dest='aiVariants',
         default=None, help='use AI variants: comma separated list',
         metavar='AI')
-    parser.add_option('', '--log', dest='log', action='store_true',
+    parser.add_option(
+        '', '--log', dest='log', action='store_true',
         default=False, help='write detailled debug info to ~/.kajongg/log/game/ruleset/commit.'
-                ' This starts a separate server process per job, it sets --servers to --clients.')
-    parser.add_option('', '--game', dest='game',
-        help='start first game with GAMEID, increment for following games.'
-            ' Without this, random values are used.',
+                            ' This starts a separate server process per job, it sets --servers to --clients.')
+    parser.add_option(
+        '', '--game', dest='game',
+        help='start first game with GAMEID, increment for following games.' +
+        ' Without this, random values are used.',
         metavar='GAMEID', type=int, default=0)
-    parser.add_option('', '--count', dest='count',
+    parser.add_option(
+        '', '--count', dest='count',
         help='play COUNT games. Default is unlimited',
         metavar='COUNT', type=int, default=999999999)
-    parser.add_option('', '--playopen', dest='playopen', action='store_true',
+    parser.add_option(
+        '', '--playopen', dest='playopen', action='store_true',
         help='all robots play with visible concealed tiles', default=False)
-    parser.add_option('', '--clients', dest='clients',
+    parser.add_option(
+        '', '--clients', dest='clients',
         help='start a maximum of CLIENTS kajongg instances. Default is 2',
         metavar='CLIENTS', type=int, default=1)
-    parser.add_option('', '--servers', dest='servers',
+    parser.add_option(
+        '', '--servers', dest='servers',
         help='start a maximum of SERVERS kajonggserver instances. Default is 1',
         metavar='SERVERS', type=int, default=1)
-    parser.add_option('', '--git', dest='git',
+    parser.add_option(
+        '', '--git', dest='git',
         help='check all commits: either a comma separated list or a range from..until')
-    parser.add_option('', '--debug', dest='debug',
+    parser.add_option(
+        '', '--debug', dest='debug',
         help=Debug.help())
-    parser.add_option(
-        '', '--client3', dest='client3', action='store_true', default=False,
-        help='use Python 3 for all clients. This will use ports instead of sockets because'
-        ' twisted does not yet support sockets for Python 3')
-    parser.add_option(
-        '', '--server3', dest='server3', action='store_true', default=False,
-        help='use Python 3 for all servers. This will use ports instead of sockets because'
-        ' twisted does not yet support sockets for Python 3')
 
     return parser.parse_args()
 
@@ -611,7 +644,7 @@ def improve_options():
     cmd = ['python', cmdPath, '--rulesets=']
     OPTIONS.knownRulesets = subprocess.Popen(
         cmd,
-        stdout=subprocess.PIPE).communicate()[0].split('\n')
+        stdout=subprocess.PIPE).communicate()[0].decode().split('\n')
     OPTIONS.knownRulesets = list(x.strip()
                                  for x in OPTIONS.knownRulesets if x.strip())
     if OPTIONS.rulesets == 'ALL':
@@ -640,14 +673,13 @@ def improve_options():
         OPTIONS.rulesets = usingRulesets
     if OPTIONS.git is not None:
         if '..' in OPTIONS.git:
-            if not '^' in OPTIONS.git:
+            if '^' not in OPTIONS.git:
                 OPTIONS.git = OPTIONS.git.replace('..', '^..')
             commits = subprocess.check_output(
                 'git log --pretty=%h {range}'.format(
-                    range=OPTIONS.git).split(
-            ))
+                    range=OPTIONS.git).split()).decode()
             OPTIONS.git = list(reversed(list(x.strip()
-                               for x in commits.split('\n') if x.strip())))
+                                             for x in commits.split('\n') if x.strip())))
         else:
             OPTIONS.git = onlyExistingCommits(OPTIONS.git.split(','))
             if not OPTIONS.git:
@@ -660,13 +692,16 @@ def improve_options():
         OPTIONS.servers = OPTIONS.clients
         OPTIONS.debug.extend(
             'neutral,dangerousGame,explain,originalCall,robbingKong,robotAI,scores,traffic,hand'.split(','))
-    if gitHead() not in ('current', None):
+    if gitHead() not in ('current', None) and not OPTIONS.log:
         OPTIONS.debug.append('git')
     if not OPTIONS.aiVariants:
         OPTIONS.aiVariants = 'Default'
     OPTIONS.allAis = OPTIONS.aiVariants.split(',')
-    print('rulesets:', ', '.join(OPTIONS.rulesets))
-    print('AIs:', ' '.join(OPTIONS.allAis))
+    if OPTIONS.count:
+        print('rulesets:', ', '.join(OPTIONS.rulesets))
+        _ = ' '.join(OPTIONS.allAis)
+        if _ != 'Default':
+            print('AIs:', _)
     if OPTIONS.git:
         print('commits:', ' '.join(OPTIONS.git))
         # since we order jobs by game, commit we want one permanent server per
@@ -674,7 +709,7 @@ def improve_options():
     OPTIONS.jobs = allJobs()
     OPTIONS.games = allGames()
     OPTIONS.jobCount = 0
-    OPTIONS.usePort = os.name == 'nt' or OPTIONS.server3 or OPTIONS.client3
+    OPTIONS.usePort = os.name == 'nt'
 
 
 def allGames():
@@ -697,13 +732,19 @@ def allJobs():
                     OPTIONS.jobCount += 1
                     if OPTIONS.jobCount > OPTIONS.count:
                         raise StopIteration
-                    yield Job(ruleset, aiVariant, commitId, game)
-
+                    if OPTIONS.py23:
+                        yield Job(2, ruleset, aiVariant, commitId, game)
+                        yield Job(3, ruleset, aiVariant, commitId, game)
+                    else:
+                        yield Job(3 if isPython3 else 2, ruleset, aiVariant, commitId, game)
 
 def main():
     """parse options, play, evaluate results"""
     global OPTIONS  # pylint: disable=global-statement
 
+    # we want only english in the logs because i18n and friends
+    # behave differently in kde and kdestub
+    os.environ['LANG'] = 'en_US.UTF-8'
     (OPTIONS, args) = parse_options()
     OPTIONS.csv = os.path.expanduser(
         os.path.join('~', '.kajongg', 'kajongg.csv'))
@@ -712,10 +753,10 @@ def main():
 
     removeInvalidCommits(OPTIONS.csv)
 
-    evaluate(readGames(OPTIONS.csv))
-
     improve_options()
 
+    evaluate(readGames(OPTIONS.csv))
+
     errorMessage = Debug.setOptions(','.join(OPTIONS.debug))
     if errorMessage:
         print(errorMessage)
@@ -727,15 +768,15 @@ def main():
 
     print()
 
-    doJobs()
-    if OPTIONS.csv:
-        evaluate(readGames(OPTIONS.csv))
+    if OPTIONS.count:
+        doJobs()
+        if OPTIONS.csv:
+            evaluate(readGames(OPTIONS.csv))
 
 
 def cleanup(sig, dummyFrame):
     """at program end"""
     Server.stopAll()
-    Clone.removeAll()
     sys.exit(sig)
 
 # is one server for two clients.
