@@ -22,10 +22,12 @@ import weakref
 from collections import defaultdict
 
 from log import logException, logWarning, m18n, m18nc, m18nE
-from common import WINDS, IntDict, Debug, unicode
+from common import IntDict, Debug, unicode
 from common import StrMixin
+from wind import East
 from query import Query
 from tile import Tile, TileList, elements
+from tilesource import TileSource
 from meld import Meld, MeldList
 from permutations import Permutations
 from message import Message
@@ -133,14 +135,14 @@ class Player(StrMixin):
         self.__name = ''
         Players.createIfUnknown(name)
         self.name = name
-        self.wind = WINDS[0]
+        self.wind = East
         self.intelligence = AIDefault(self)
         self.visibleTiles = IntDict(game.visibleTiles) if game else IntDict()
         self.handCache = {}
         self.cacheHits = 0
         self.cacheMisses = 0
+        self.__lastSource = TileSource.East14th
         self.clearHand()
-        self.__lastSource = '1'  # no source: blessing from heaven or earth
         self.handBoard = None
 
     def __lt__(self, other):
@@ -174,7 +176,7 @@ class Player(StrMixin):
         """write once"""
         assert self.__name == ''
         assert value
-        assert isinstance(value, unicode)
+        assert isinstance(value, unicode), 'Player.name must be unicode but not {}'.format(type(value))
         self.__name = value
 
     @property
@@ -194,7 +196,7 @@ class Player(StrMixin):
         self.newHandContent = None
         self.originalCallingHand = None
         self.__lastTile = None
-        self.lastSource = '1'
+        self.lastSource = TileSource.East14th
         self.lastMeld = Meld()
         self.__mayWin = True
         self.__payment = 0
@@ -224,10 +226,12 @@ class Player(StrMixin):
 
     @property
     def hand(self):
-        """a readonly tuple"""
-# TODO: str or what?
+        """readonly: the current Hand. Compute if invalidated."""
         if not self._hand:
-            self._hand = self.computeHand()
+            self._hand = self.__computeHand()
+        elif Debug.hand:
+            _ = self.__computeHand()
+            assert self._hand == self.__computeHand(), '{} != {}'.format(_, self._hand)
         return self._hand
 
     @property
@@ -268,13 +272,15 @@ class Player(StrMixin):
         return self.__lastSource
 
     @lastSource.setter
-    def lastSource(self, lastSource):
+    def lastSource(self, value):
         """the source of the last tile the player got"""
-        self.__lastSource = lastSource
-        if lastSource == 'd' and not self.game.wall.living:
-            self.__lastSource = 'Z'
-        if lastSource == 'w' and not self.game.wall.living:
-            self.__lastSource = 'z'
+        if value is TileSource.LivingWallDiscard and not self.game.wall.living:
+            value = TileSource.LivingWallEndDiscard
+        if value is TileSource.LivingWall and not self.game.wall.living:
+            value = TileSource.LivingWallEnd
+        if self.__lastSource != value:
+            self.__lastSource = value
+            self._hand = None
 
     @property
     def nameid(self):
@@ -334,10 +340,10 @@ class Player(StrMixin):
             self.lastTile = tile
         self.addConcealedTiles([tile])
         if deadEnd:
-            self.lastSource = 'e'
+            self.lastSource = TileSource.DeadWall
         else:
             self.game.lastDiscard = None
-            self.lastSource = 'w'
+            self.lastSource = TileSource.LivingWall
         return self.lastTile
 
     def removeTile(self, tile):
@@ -358,7 +364,6 @@ class Player(StrMixin):
         """add to my tiles"""
         assert len(tiles)
         for tile in tiles:
-            assert isinstance(tile, Tile), 'tile:%s' % tile
             if tile.isBonus:
                 self._bonusTiles.append(tile)
             else:
@@ -380,8 +385,8 @@ class Player(StrMixin):
 
     def mjString(self):
         """compile hand info into a string as needed by the scoring engine"""
-        declaration = 'a' if self.originalCall else ''
-        return ''.join(['m', self.lastSource, declaration])
+        announcements = 'a' if self.originalCall else ''
+        return ''.join(['m', self.lastSource.char, ''.join(announcements)])
 
     def makeTileKnown(self, tileName):
         """used when somebody else discards a tile"""
@@ -389,26 +394,37 @@ class Player(StrMixin):
         self._concealedTiles[0] = tileName
         self._hand = None
 
-    def computeHand(self, withTile=None):
+    def __computeHand(self):
         """returns Hand for this player"""
         assert not (self._concealedMelds and self._concealedTiles)
-        assert isinstance(self.lastTile, (Tile, type(None)))
-        assert isinstance(withTile, (Tile, type(None)))
-        melds = ['R' + ''.join(str(x) for x in sorted(self._concealedTiles))]
-        if withTile:
-            melds[0] += withTile
-        if melds[0] == 'R':
-            melds = melds[1:]
+        melds = list()
         melds.extend(str(x) for x in self._exposedMelds)
         melds.extend(str(x) for x in self._concealedMelds)
+        if self._concealedTiles:
+            melds.append('R' + ''.join(str(x) for x in sorted(self._concealedTiles)))
         melds.extend(str(x) for x in self._bonusTiles)
         melds.append(self.mjString())
-        if withTile or self.lastTile:
+        if self.lastTile:
             melds.append(
                 'L%s%s' %
-                (withTile or self.lastTile, self.lastMeld if self.lastMeld else ''))
+                (self.lastTile, self.lastMeld if self.lastMeld else ''))
         return Hand(self, ' '.join(melds))
 
+    def _computeHandWithDiscard(self, discard):
+        """what if"""
+        lastSource = self.lastSource # TODO: recompute
+        save = (self.lastTile, self.lastSource)
+        try:
+            self.lastSource = lastSource
+            if discard:
+                self.lastTile = discard
+                self._concealedTiles.append(discard)
+            return self.__computeHand()
+        finally:
+            self.lastTile, self.lastSource = save
+            if discard:
+                self._concealedTiles = self._concealedTiles[:-1]
+
     def scoringString(self):
         """helper for HandBoard.__str__"""
         if self._concealedMelds:
@@ -480,18 +496,22 @@ class PlayingPlayer(Player):
 
     def declaredMahJongg(self, concealed, withDiscard, lastTile, lastMeld):
         """player declared mah jongg. Determine last meld, show concealed tiles grouped to melds"""
+        if Debug.mahJongg:
+            self.game.debug('{} declared MJ: concealed={}, withDiscard={}, lastTile={},lastMeld={}'.format(
+                self, concealed, withDiscard, lastTile, lastMeld))
+            self.game.debug('  with hand being {}'.format(self.hand))
         melds = concealed[:]
         self.game.winner = self
+        assert lastMeld in melds, \
+            'lastMeld %s not in melds: concealed=%s: melds=%s lastTile=%s withDiscard=%s' % (
+                lastMeld, self._concealedTiles, melds, lastTile, withDiscard)
         if withDiscard:
-            assert isinstance(withDiscard, Tile), withDiscard
             PlayingPlayer.addConcealedTiles(
                 self,
                 [withDiscard])  # this should NOT invoke syncHandBoard
-            if self.lastSource != 'k':   # robbed the kong
-                self.lastSource = 'd'
+            if self.lastSource is not TileSource.RobbedKong:
+                self.lastSource = TileSource.LivingWallDiscard
             # the last claimed meld is exposed
-            assert lastMeld in melds, '%s: melds=%s lastMeld=%s lastTile=%s withDiscard=%s' % (
-                self._concealedTiles, melds, lastMeld, lastTile, withDiscard)
             melds.remove(lastMeld)
             lastTile = withDiscard.exposed
             lastMeld = lastMeld.exposed
@@ -503,6 +523,9 @@ class PlayingPlayer(Player):
         self._concealedMelds = melds
         self._concealedTiles = []
         self._hand = None
+        if Debug.mahJongg:
+            self.game.debug('  hand becomes {}'.format(self.hand))
+            self._hand = None
 
     def __possibleChows(self):
         """returns a unique list of lists with possible claimable chow combinations"""
@@ -521,7 +544,7 @@ class PlayingPlayer(Player):
         kongs = []
         if self == self.game.activePlayer:
             # declaring a kong
-            for tileName in set([x for x in self._concealedTiles if not x.isBonus]):
+            for tileName in sorted(set([x for x in self._concealedTiles if not x.isBonus])):
                 if self._concealedTiles.count(tileName) == 4:
                     kongs.append(tileName.kong)
                 elif self._concealedTiles.count(tileName) == 1 and \
@@ -533,17 +556,17 @@ class PlayingPlayer(Player):
             # claiming a kong
             discardTile = self.game.lastDiscard.concealed
             if self._concealedTiles.count(discardTile) == 3:
-                # TODO: discard.kong.concealed is aAAa but we need AAAA
+                # discard.kong.concealed is aAAa but we need AAAA
                 kongs.append(Meld(discardTile * 4))
         return kongs
 
-    def __maySayChow(self):
+    def __maySayChow(self, dummyMove):
         """returns answer arguments for the server if calling chow is possible.
         returns the meld to be completed"""
         if self == self.game.nextPlayer():
             return self.__possibleChows()
 
-    def __maySayPung(self):
+    def __maySayPung(self, dummyMove):
         """returns answer arguments for the server if calling pung is possible.
         returns the meld to be completed"""
         lastDiscard = self.game.lastDiscard
@@ -552,7 +575,7 @@ class PlayingPlayer(Player):
             if self.concealedTiles.count(lastDiscard) >= 2:
                 return MeldList([lastDiscard.pung])
 
-    def __maySayKong(self):
+    def __maySayKong(self, dummyMove):
         """returns answer arguments for the server if calling or declaring kong is possible.
         returns the meld to be completed or to be declared"""
         return self.__possibleKongs()
@@ -566,7 +589,7 @@ class PlayingPlayer(Player):
             withDiscard = game.lastDiscard
         else:
             withDiscard = None
-        hand = self.computeHand(withTile=withDiscard)
+        hand = self._computeHandWithDiscard(withDiscard)
         if hand.won:
             if Debug.robbingKong:
                 if move.message == Message.DeclaredKong:
@@ -575,11 +598,12 @@ class PlayingPlayer(Player):
             if Debug.mahJongg:
                 game.debug('%s may say MJ:%s, active=%s' % (
                     self, list(x for x in game.players), game.activePlayer))
+                game.debug('  with hand {}'.format(hand))
             return MeldList(x for x in hand.melds if not x.isDeclared), withDiscard, hand.lastMeld
 
-    def __maySayOriginalCall(self):
+    def __maySayOriginalCall(self, dummyMove):
         """returns True if Original Call is possible"""
-        for tileName in set(self.concealedTiles):
+        for tileName in sorted(set(self.concealedTiles)):
             newHand = self.hand - tileName
             if newHand.callingHands:
                 if Debug.originalCall:
@@ -588,21 +612,21 @@ class PlayingPlayer(Player):
                         (self, tileName, self.hand))
                 return True
 
+    sayables = {
+        Message.Pung: __maySayPung,
+        Message.Kong: __maySayKong,
+        Message.Chow: __maySayChow,
+        Message.MahJongg: __maySayMahjongg,
+        Message.OriginalCall: __maySayOriginalCall}
+
     def computeSayable(self, move, answers):
         """find out what the player can legally say with this hand"""
         self.sayable = {}
         for message in Message.defined.values():
-            self.sayable[message] = True
-        if Message.Pung in answers:
-            self.sayable[Message.Pung] = self.__maySayPung()
-        if Message.Chow in answers:
-            self.sayable[Message.Chow] = self.__maySayChow()
-        if Message.Kong in answers:
-            self.sayable[Message.Kong] = self.__maySayKong()
-        if Message.MahJongg in answers:
-            self.sayable[Message.MahJongg] = self.__maySayMahjongg(move)
-        if Message.OriginalCall in answers:
-            self.sayable[Message.OriginalCall] = self.__maySayOriginalCall()
+            if message in answers and message in self.sayables:
+                self.sayable[message] = self.sayables[message](self, move)
+            else:
+                self.sayable[message] = True
 
     def maybeDangerous(self, msg):
         """could answering with msg lead to dangerous game?
@@ -637,7 +661,7 @@ class PlayingPlayer(Player):
                     tileName, Tile.unknown)
                 assert src != dst, (
                     self, src, dst, tiles, self._concealedTiles)
-                if not src in self._concealedTiles:
+                if src not in self._concealedTiles:
                     logException('%s: showConcealedTiles(%s): %s not in %s.' %
                                  (self, tiles, src, self._concealedTiles))
                 idx = self._concealedTiles.index(src)
@@ -655,7 +679,7 @@ class PlayingPlayer(Player):
                 if tile == ignoreDiscard:
                     ignoreDiscard = None
                 else:
-                    if not tile in self._concealedTiles:
+                    if tile not in self._concealedTiles:
                         msg = m18nE(
                             '%1 claiming MahJongg: She does not really have tile %2')
                         return msg, self.name, tile
@@ -670,8 +694,8 @@ class PlayingPlayer(Player):
             return msg, self.name
         self._hand = None
 
-    def robTile(self, tile):
-        """used for robbing the kong"""
+    def robTileFrom(self, tile):
+        """used for robbing the kong from this player"""
         assert tile.isConcealed
         tile = tile.exposed
         for meld in self._exposedMelds:
@@ -680,11 +704,15 @@ class PlayingPlayer(Player):
                 self.visibleTiles[tile] -= 1
                 break
         else:
-            raise Exception('robTile: no meld found with %s' % tile)
+            raise Exception('robTileFrom: no meld found with %s' % tile)
         self.game.lastDiscard = tile.concealed
         self.lastTile = None  # our lastTile has just been robbed
         self._hand = None
 
+    def robsTile(self):
+        """True if the player is robbing a tile"""
+        self.lastSource = TileSource.RobbedKong
+
     def scoreMatchesServer(self, score):
         """do we compute the same score as the server does?"""
         if score is None:
