@@ -1,3 +1,4 @@
+#pylint: disable=too-many-lines
 # -*- coding: utf-8 -*-
 
 """Copyright (C) 2009-2014 Wolfgang Rohdewald <wolfgang@rohdewald.de>
@@ -18,8 +19,10 @@ Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA 02110-1301, USA.
 """
 
 from tile import Tile, elements
+from tilesource import TileSource
 from meld import Meld, MeldList
-from common import IntDict, WINDS
+from common import IntDict
+from wind import East
 from message import Message
 from query import Query
 from permutations import Permutations
@@ -39,10 +42,17 @@ class RuleCode(object):
 
     All methods in RuleCode classes will automatically be converted
     into staticmethods or classmethods if the 1st arg is named 'cls'.
+
+    winningTileCandidates(cls, hand):
+        All rules for going MahJongg must have such a method.
+        This is used to find all winning hands which only need
+        one tile: The calling hands (after calling)
+
     """
 
     cache = ()
 
+
 # pylint: disable=missing-docstring
 # the class and method names are mostly self explaining, we do not
 # need docstringss
@@ -50,6 +60,11 @@ class RuleCode(object):
 # pylint: disable=no-self-argument, no-self-use, no-value-for-parameter, no-member
 # pylint: disable=too-many-function-args, unused-argument, arguments-differ
 
+class MJRule(RuleCode):
+
+    def computeLastMelds(hand):
+        """returns all possible last melds"""
+
 
 class DragonPungKong(RuleCode):
 
@@ -132,7 +147,7 @@ class ConcealedHonorsKong(RuleCode):
 class OwnWindPungKong(RuleCode):
 
     def appliesToMeld(hand, meld):
-        return meld[0].value == hand.ownWind
+        return meld[0].value is hand.ownWind
 
     def mayApplyToMeld(meld):
         """for meld rules which depend on context like hand.ownWind, we want
@@ -146,7 +161,7 @@ class OwnWindPungKong(RuleCode):
 class OwnWindPair(RuleCode):
 
     def appliesToMeld(hand, meld):
-        return meld[0].value == hand.ownWind
+        return meld[0].value is hand.ownWind
 
     def mayApplyToMeld(meld):
         return meld.isPair and meld.isWindMeld
@@ -155,7 +170,7 @@ class OwnWindPair(RuleCode):
 class RoundWindPungKong(RuleCode):
 
     def appliesToMeld(hand, meld):
-        return meld[0].value == hand.roundWind
+        return meld[0].value is hand.roundWind
 
     def mayApplyToMeld(meld):
         return meld.isPungKong and meld.isWindMeld
@@ -164,7 +179,7 @@ class RoundWindPungKong(RuleCode):
 class RoundWindPair(RuleCode):
 
     def appliesToMeld(hand, meld):
-        return meld[0].value == hand.roundWind
+        return meld[0].value is hand.roundWind
 
     def mayApplyToMeld(meld):
         return meld.isPair and meld.isWindMeld
@@ -247,7 +262,7 @@ class Purity(RuleCode):
 class ConcealedTrueColorGame(RuleCode):
 
     def appliesToHand(hand):
-        if len(hand.suits) != 1 or not hand.suits < set(Tile.colors):
+        if len(hand.suits) != 1 or hand.suits >= set(Tile.colors):
             return False
         return not any((x.isExposed and not x.isClaimedKong) for x in hand.melds)
 
@@ -286,7 +301,8 @@ class AllTerminals(RuleCode):
         return all(x.isTerminal for x in hand.tiles)
 
 
-class StandardMahJongg(RuleCode):
+class StandardMahJongg(MJRule):
+
     cache = ('appliesToHand',)
 
     def computeLastMelds(hand):
@@ -331,8 +347,7 @@ class StandardMahJongg(RuleCode):
             return {Tile(group, val0 + 1)}
 
     def winningTileCandidates(cls, hand):
-        # pylint:
-        # disable=too-many-locals,too-many-return-statements,too-many-branches,too-many-statements
+        # pylint: disable=too-many-locals,too-many-return-statements,too-many-branches,too-many-statements
         if len(hand.melds) > 7:
             # hope 7 is sufficient, 6 was not
             return set()
@@ -371,17 +386,19 @@ class StandardMahJongg(RuleCode):
         if maxChows == 0:
             return set(result)
         melds = []
-        for group in hand.suits & set(Tile.colors):
+        for group in sorted(hand.suits & set(Tile.colors)):
             values = sorted(x.value for x in result if x.group == group)
             changed = True
             while (changed and len(values) > 2
-                    and values.count(values[0]) == 1
-                    and values.count(values[1]) == 1
-                    and values.count(values[2]) == 1):
+                   and values.count(values[0]) == 1
+                   and values.count(values[1]) == 1
+                   and values.count(values[2]) == 1):
                 changed = False
                 if values[0] + 2 == values[2] and (len(values) == 3 or values[3] > values[0] + 3):
                     # logDebug(u'removing first 3 from %s' % values)
                     meld = Tile(group, values[0]).chow
+                    # pylint: disable=not-an-iterable
+                    # must be a pylint bug. meld is TileList is list
                     for pair in meld:
                         result.remove(pair)
                     melds.append(meld)
@@ -393,12 +410,14 @@ class StandardMahJongg(RuleCode):
                     return cls.fillChow(group, values[:2])
             changed = True
             while (changed and len(values) > 2
-                    and values.count(values[-1]) == 1
-                    and values.count(values[-2]) == 1
-                    and values.count(values[-3]) == 1):
+                   and values.count(values[-1]) == 1
+                   and values.count(values[-2]) == 1
+                   and values.count(values[-3]) == 1):
                 changed = False
                 if values[-1] - 2 == values[-3] and (len(values) == 3 or values[-4] < values[-1] - 3):
                     meld = Tile(group, values[-3]).chow
+                    # pylint: disable=not-an-iterable
+                    # must be a pylint bug. meld is TileList is list
                     for pair in meld:
                         result.remove(pair)
                     melds.append(meld)
@@ -475,7 +494,7 @@ class SquirmingSnake(StandardMahJongg):
         std = hand.ruleCache.get(cacheKey, None)
         if std is False:
             return False
-        if len(hand.suits) != 1 or not hand.suits < set(Tile.colors):
+        if len(hand.suits) != 1 or hand.suits >= set(Tile.colors):
             return False
         values = hand.values
         if values.count(1) < 3 or values.count(9) < 3:
@@ -490,7 +509,7 @@ class SquirmingSnake(StandardMahJongg):
         return set()
 
 
-class WrigglingSnake(RuleCode):
+class WrigglingSnake(MJRule):
 
     def shouldTry(hand, maxMissing=3):
         if hand.declaredMelds:
@@ -544,7 +563,7 @@ class WrigglingSnake(RuleCode):
         if Tile.wind not in suits:
             return False
         suits -= {Tile.wind}
-        if len(suits) != 1 or not suits < set(Tile.colors):
+        if len(suits) != 1 or suits >= set(Tile.colors):
             return False
         if hand.values.count(1) != 2:
             return False
@@ -561,7 +580,7 @@ class CallingHand(RuleCode):
         return False
 
 
-class TripleKnitting(RuleCode):
+class TripleKnitting(MJRule):
 
     def computeLastMelds(cls, hand):
         """returns all possible last melds"""
@@ -599,7 +618,8 @@ class TripleKnitting(RuleCode):
             rest.remove(triple[1])
             rest.remove(triple[2])
         while len(rest) >= 2:
-            for value in set(x.value for x in rest):
+            for tile in sorted(set(rest)):
+                value = tile.value
                 suits = set(x.group for x in rest if x.value == value)
                 if len(suits) < 2:
                     yield tuple(melds), tuple(rest)
@@ -638,7 +658,7 @@ class TripleKnitting(RuleCode):
     def shouldTry(cls, hand, maxMissing=3):
         if hand.declaredMelds:
             return False
-        tripleWanted = 7 - maxMissing // 3  # count triples
+        tripleWanted = 4 - maxMissing // 3  # count triples
         tripleCount = len(cls.findTriples(hand)[0])
         return tripleCount >= tripleWanted
 
@@ -666,7 +686,7 @@ class TripleKnitting(RuleCode):
         return tuple(result), tuple(tilesS + tilesB + tilesC)
 
 
-class Knitting(RuleCode):
+class Knitting(MJRule):
 
     def computeLastMelds(cls, hand):
         """returns all possible last melds"""
@@ -765,16 +785,16 @@ class Knitting(RuleCode):
     def pairSuits(hand):
         """returns a lowercase string with two suit characters. If no prevalence, returns None"""
         suitCounts = list(len([x for x in hand.tiles if x.lowerGroup == y])
-                          for y in set(Tile.colors))
+                          for y in Tile.colors)
         minSuit = min(suitCounts)
         result = ''.join(
             x for idx,
-            x in enumerate(set(Tile.colors)) if suitCounts[idx] > minSuit)
+            x in enumerate(Tile.colors) if suitCounts[idx] > minSuit)
         if len(result) == 2:
             return result
 
 
-class AllPairHonors(RuleCode):
+class AllPairHonors(MJRule):
 
     def computeLastMelds(hand):
         return [hand.lastTile.pair]
@@ -798,7 +818,7 @@ class AllPairHonors(RuleCode):
         if len(set(hand.tiles)) != 7:
             return False
         tileCounts = list([len([x for x in hand.tiles if x == y])
-                          for y in hand.tiles])
+                           for y in hand.tiles])
         return set(tileCounts) == set([2])
 
     def winningTileCandidates(cls, hand):
@@ -827,7 +847,7 @@ class AllPairHonors(RuleCode):
 
     def rearrange(hand, rest):
         melds = []
-        for pair in set(rest) & elements.mAJORS:
+        for pair in sorted(set(rest) & elements.mAJORS):
             while rest.count(pair) >= 2:
                 melds.append(pair.pair)
                 rest.remove(pair)
@@ -886,7 +906,7 @@ class LittleThreeDragons(RuleCode):
 
     def appliesToHand(hand):
         lengths = sorted([min(len(x), 3)
-                         for x in hand.melds if x.isDragonMeld])
+                          for x in hand.melds if x.isDragonMeld])
         return lengths == [2, 3, 3]
 
 
@@ -905,47 +925,47 @@ class AllGreen(RuleCode):
 class LastTileFromWall(RuleCode):
 
     def appliesToHand(hand):
-        return hand.lastSource == 'w'
+        return hand.lastSource is TileSource.LivingWall
 
 
 class LastTileFromDeadWall(RuleCode):
 
     def appliesToHand(hand):
-        return hand.lastSource == 'e'
+        return hand.lastSource is TileSource.DeadWall
 
     def selectable(hand):
         """for scoring game"""
-        return hand.lastSource == 'w'
+        return hand.lastSource is TileSource.LivingWall
 
 
 class IsLastTileFromWall(RuleCode):
 
     def appliesToHand(hand):
-        return hand.lastSource == 'z'
+        return hand.lastSource is TileSource.LivingWallEnd
 
     def selectable(hand):
         """for scoring game"""
-        return hand.lastSource == 'w'
+        return hand.lastSource is TileSource.LivingWall
 
 
 class IsLastTileFromWallDiscarded(RuleCode):
 
     def appliesToHand(hand):
-        return hand.lastSource == 'Z'
+        return hand.lastSource is TileSource.LivingWallEndDiscard
 
     def selectable(hand):
         """for scoring game"""
-        return hand.lastSource == 'd'
+        return hand.lastSource is TileSource.LivingWallDiscard
 
 
 class RobbingKong(RuleCode):
 
     def appliesToHand(hand):
-        return hand.lastSource == 'k'
+        return hand.lastSource is TileSource.RobbedKong
 
     def selectable(hand):
         """for scoring game"""
-        return (hand.lastSource and hand.lastSource in 'kwd'
+        return (hand.lastSource in (TileSource.RobbedKong, TileSource.LivingWall, TileSource.LivingWallDiscard)
                 and hand.lastTile and hand.lastTile.group.islower()
                 and [x.exposed for x in hand.tiles].count(hand.lastTile.exposed) < 2)
 
@@ -971,7 +991,7 @@ class ScratchingPole(RuleCode):
 class StandardRotation(RuleCode):
 
     def rotate(game):
-        return game.winner and game.winner.wind != 'E'
+        return game.winner and game.winner.wind is not East
 
 
 class EastWonNineTimesInARow(RuleCode):
@@ -986,12 +1006,11 @@ class EastWonNineTimesInARow(RuleCode):
             if game.isScoringGame():
                 # we are only proposing for the last needed Win
                 needWins -= 1
-        if game.winner and game.winner.wind == 'E' and game.notRotated >= needWins:
-            prevailing = WINDS[game.roundsFinished % 4]
+        if game.winner and game.winner.wind is East and game.notRotated >= needWins:
             eastMJCount = int(Query("select count(1) from score "
                                     "where game=%d and won=1 and wind='E' and player=%d "
                                     "and prevailing='%s'" %
-                                    (game.gameid, game.players['E'].nameid, prevailing)).records[0][0])
+                                    (game.gameid, game.players[East].nameid, game.roundWind.char)).records[0][0])
             return eastMJCount == needWins
         return False
 
@@ -1000,14 +1019,19 @@ class EastWonNineTimesInARow(RuleCode):
 
 
 class GatesOfHeaven(StandardMahJongg):
+    """as used for Classical Chinese BMJA.
+
+    I believe that when they say a run of 2..8, they must
+    all be concealed
+    """
+
     cache = ()
 
-# FIXME: computeLastMelds must return a real meld
-# FIXME: in BMJA, 111 and 999 must be concealed, we do not check this
+# TODO: in BMJA, 111 and 999 must be concealed, we do not check this
     def computeLastMelds(hand):
-        return [hand.lastTile.single]
+        return [x for x in hand.melds if hand.lastTile in x]
 
-    def shouldTry(hand, maxMissing=3):
+    def shouldTry(hand, maxMissing=None):
         if hand.declaredMelds:
             return False
         for suit in Tile.colors:
@@ -1018,13 +1042,10 @@ class GatesOfHeaven(StandardMahJongg):
                 return True
         return False
 
-    def maybeCallingOrWon(hand):
-        if len(hand.suits) != 1 or not hand.suits < set(Tile.colors):
-            return False
-        return not hand.declaredMelds
-
     def appliesToHand(cls, hand):
-        if not cls.maybeCallingOrWon(hand):
+        if len(hand.suits) != 1 or hand.suits >= set(Tile.colors):
+            return False
+        if any(len(x) > 2 for x in hand.declaredMelds):
             return False
         values = hand.values
         if len(set(values)) < 9 or values.count(1) != 3 or values.count(9) != 3:
@@ -1036,35 +1057,83 @@ class GatesOfHeaven(StandardMahJongg):
         if len(values) != 1:
             return False
         surplus = values[0]
-        if 'pair28' in cls.options:
-            return 1 < surplus < 9
-        if 'lastExtra' in cls.options:
-            return hand.lastTile and surplus == hand.lastTile.value
-        return True
+        return 1 < surplus < 9
 
     def winningTileCandidates(cls, hand):
-        result = set()
-        if not cls.maybeCallingOrWon(hand):
-            return result
+        if hand.declaredMelds:
+            return set()
+        if len(hand.suits) != 1 or hand.suits >= set(Tile.colors):
+            return set()
         values = hand.values
-        if len(set(values)) == 8:
-            # one minor is missing
-            result = set(Tile.minors) - set(values)
-        else:
-            # we have something of all values
-            if values.count(1) != 3:
-                result = (1,)
-            elif values.count(9) != 3:
-                result = (9,)
-            else:
-                if 'pair28' in cls.options:
-                    result = Tile.minors
-                else:
-                    result = Tile.numbers
-        return {Tile(list(hand.suits)[0], x) for x in result}
+        if len(set(values)) < 9:
+            return set()
+        # we have something of all values
+        if values.count(1) != 3 or values.count(9) != 3:
+# TODO: we may get them from the wall but not by claim. Differentiate!
+            return set()
+        for suit in hand.suits:
+            return {Tile(suit, x) for x in Tile.minors}
+
+    def rearrange(hand, rest):
+        melds = []
+        for suit in hand.suits & set(Tile.colors):
+            for value in Tile.numbers:
+                tile = Tile(suit, value).concealed
+                if rest.count(tile) == 3 and tile.isTerminal:
+                    melds.append(tile.pung)
+                    rest.remove(tile)
+                    rest.remove(tile)
+                    rest.remove(tile)
+                elif rest.count(tile) == 2:
+                    melds.append(tile.pair)
+                    rest.remove(tile)
+                    rest.remove(tile)
+                elif rest.count(tile) == 1:
+                    melds.append(tile.single)
+                    rest.remove(tile)
+            break
+        yield tuple(melds), tuple(rest)
 
 
-class ThirteenOrphans(RuleCode):
+class NineGates(GatesOfHeaven):
+    """as used for Classical Chinese DMJL"""
+
+    def appliesToHand(cls, hand):
+        """last tile may also be 1 or 9"""
+        if hand.declaredMelds:
+            return False
+        if len(hand.suits) != 1 or hand.suits >= set(Tile.colors):
+            return False
+        values = hand.values
+        if len(set(values)) < 9:
+            return False
+        if values.count(1) != 3 or values.count(9) != 3:
+            return False
+        values = list(values[3:-3])
+        for value in Tile.minors:
+            if value in values:
+                values.remove(value)
+        if len(values) != 1:
+            return False
+        surplus = values[0]
+        return hand.lastTile and surplus == hand.lastTile.value
+
+    def winningTileCandidates(cls, hand):
+        if hand.declaredMelds:
+            return set()
+        if len(hand.suits) != 1 or hand.suits >= set(Tile.colors):
+            return set()
+        values = hand.values
+        if len(set(values)) < 9:
+            return set()
+        # we have something of all values
+        if values.count(1) != 3 or values.count(9) != 3:
+            return set()
+        for suit in hand.suits:
+            return {Tile(suit, x) for x in Tile.numbers}
+
+
+class ThirteenOrphans(MJRule):
 
     def computeLastMelds(hand):
         meldSize = hand.tilesInHand.count(hand.lastTile)
@@ -1151,25 +1220,29 @@ class ThirteenOrphans(RuleCode):
 class OwnFlower(RuleCode):
 
     def appliesToMeld(hand, meld):
-        return meld[0].value == hand.ownWind
+        return meld[0].value is hand.ownWind
 
     def mayApplyToMeld(meld):
+        # pylint: disable=unsubscriptable-object
+        # must be a pylint bug. meld is TileList is list
         return meld.isBonus and meld[0].group == Tile.flower
 
 
 class OwnSeason(RuleCode):
 
     def appliesToMeld(hand, meld):
-        return meld[0].value == hand.ownWind
+        return meld[0].value is hand.ownWind
 
     def mayApplyToMeld(meld):
+        # pylint: disable=unsubscriptable-object
+        # must be a pylint bug. meld is TileList is list
         return meld.isBonus and meld[0].group == Tile.season
 
 
 class OwnFlowerOwnSeason(RuleCode):
 
     def appliesToHand(hand):
-        return sum(x.isBonus and x[0].value == hand.ownWind for x in hand.bonusMelds) == 2
+        return sum(x.isBonus and x[0].value is hand.ownWind for x in hand.bonusMelds) == 2
 
 
 class AllFlowers(RuleCode):
@@ -1231,25 +1304,35 @@ class TwofoldFortune(RuleCode):
 class BlessingOfHeaven(RuleCode):
 
     def appliesToHand(hand):
-        return hand.ownWind == Tile.east and hand.lastSource == '1'
+        if hand.lastSource is not TileSource.East14th:
+            return False
+        if hand.ownWind is not East:
+            return False
+        assert hand.lastTile.isConcealed, '{}: Blessing of Heaven: last tile must be concealed'.format(hand)
+        return True
 
     def selectable(hand):
         """for scoring game"""
-        return (hand.ownWind == Tile.east
-                and hand.lastSource and hand.lastSource in 'wd'
-                and not set(hand.announcements) - {'a'})
+        return (hand.ownWind is East
+                and hand.lastSource in (TileSource.LivingWall, TileSource.LivingWallDiscard)
+                and not hand.announcements - {'a'})
 
 
 class BlessingOfEarth(RuleCode):
 
     def appliesToHand(hand):
-        return hand.ownWind != Tile.east and hand.lastSource == '1'
+        if hand.lastSource is not TileSource.East14th:
+            return False
+        if hand.ownWind is East:
+            return False
+        assert hand.lastTile.isExposed, '{}: Blessing of Earth: last tile must be exposed'.format(hand)
+        return True
 
     def selectable(hand):
         """for scoring game"""
-        return (hand.ownWind != Tile.east
-                and hand.lastSource and hand.lastSource in 'wd'
-                and not set(hand.announcements) - {'a'})
+        return (hand.ownWind is not East
+                and hand.lastSource in (TileSource.LivingWall, TileSource.LivingWallDiscard)
+                and not hand.announcements - {'a'})
 
 
 class LongHand(RuleCode):
